/*
* Copyright 2001-2015 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2012 S3 Graphics, Inc. All Rights Reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a
* copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sub license,
* and/or sell copies of the Software, and to permit persons to whom the
* Software is furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice (including the
* next paragraph) shall be included in all copies or substantial portions
* of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
* THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
* OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
* ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
* DEALINGS IN THE SOFTWARE.
*/

#include "s3.h"
#include "os_interface.h"
#include "s3g_driver.h"
#include "s3g_ioctl.h"
#include "s3g_version.h"

#ifdef __mips__
extern int s3g_swfb;
#else
extern int s3g_swfb;
#endif
extern char *s3g_fb_mode;
extern int s3g_fb;
extern int s3g_flip;

extern s3g_card_t s3g_cards[S3G_MAX_CARDS];

static int s3g_open(struct inode *inode, struct file *filp)
{
    s3g_card_t *s3g = NULL;

    s3g_file_t *priv;

    int minor;

    int err = -ENODEV;

#ifndef MINOR
    minor = minor(inode->i_rdev);
#else
    minor = MINOR(inode->i_rdev);
#endif

    if(!((minor >= 0) && (minor < S3G_MAX_CARDS)))
    {
        return err;
    }

#ifdef S3G_PCIE_BUS
    s3g = &s3g_cards[minor];
#else
    s3g = &s3g_cards[0];
#endif

    if(!s3g->adapter)
    {
        return err;
    }
  
    priv = s3g_calloc(sizeof(s3g_file_t));

    if(!priv)
    {
        return -ENOMEM;
    }

    filp->private_data = priv;

    priv->filp        = filp;

    priv->card        = s3g;

    init_waitqueue_head(&priv->wait_queue);

    return 0;
}

static int s3g_release(struct inode *inode, struct file *filp)
{
    s3g_file_t *priv = filp->private_data;
    s3g_card_t *s3g  = priv->card;

    if(priv->wait_events)
    {
        s3g_remove_wait(s3g, priv);
    }

    if(priv->hold_lock)
    {
        s3g_mutex_unlock(s3g->lock);
    }

    if(priv->gpu_device)
    {
        krnl_final_cleanup(s3g->adapter, priv->gpu_device);
        priv->gpu_device = 0;
    }

    s3g_free(priv);
    filp->private_data = NULL;

    return 0;
}

static long s3g_unlocked_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    unsigned int nr   = S3G_IOCTL_NR(cmd);
    s3g_ioctl_t  func = s3g_ioctls[nr];
    s3g_ioctl_t_2  func2 = (s3g_ioctl_t_2)s3g_ioctls[nr];

    s3g_file_t *priv = filp->private_data;
    long ret;

    if(nr == (S3G_IOCTL_NR(S3G_IOCTL_LOCK_ALLOCATION)))
    {
        priv->filp = filp;
        ret = (long)func2(filp->private_data, cmd, arg, filp);
    }
    else if(nr == (S3G_IOCTL_NR(S3G_IOCTL_ESCAPE_CALL)))
    {
        priv->filp = filp;
        ret = (long)func2(filp->private_data, cmd, arg, filp);
    }
    else if(nr == (S3G_IOCTL_NR(S3G_IOCTL_2D_INIT)))
    {
          priv->filp = filp;
          ret = (long)func2(filp->private_data, cmd, arg, filp);
    }
    else if(nr == (S3G_IOCTL_NR(S3G_IOCTL_LEAVE_VT)))
    {
          priv->filp = filp;
          ret = (long)func2(filp->private_data, cmd, arg, filp);
    }
    else if(nr == (S3G_IOCTL_NR(S3G_IOCTL_ENTER_VT)))
    {
          priv->filp = filp;
          ret = (long)func2(filp->private_data, cmd, arg, filp);
    }
    else
    {


        ret = (long)func(filp->private_data, cmd, arg);
    }

    return ret;
}


static int s3g_mmap(struct file *filp, struct vm_area_struct *vma)
{
    int ret = 0;
#ifndef __frv__
    s3g_file_t     *priv = filp->private_data;
    s3g_card_t     *card = priv->card;
    s3g_map_argu_t *map  = priv->map;
    s3g_map_argu_t map_argu;

    if(map == NULL)
    {
        bus_config_t bus_config = {0};

        unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;

        s3g_get_bus_config(card->pdev, &bus_config);

        map_argu.flags.mem_type   = S3G_SYSTEM_IO;
  
        if((offset >= bus_config.start_addr[1]) && 
           (offset <  bus_config.end_addr[1]))
        {
            /* map fb as wc */
            map_argu.flags.cache_type = S3G_MEM_WRITE_COMBINED;
        }
        else
        {
            /* map mmio reg as uc*/
            map_argu.flags.cache_type = S3G_MEM_UNCACHED;
        }

        map = &map_argu;
    }

    switch(map->flags.mem_type)
    {
        case S3G_SYSTEM_IO:
            ret = s3g_map_system_io(vma, map);
            break;
            
        case S3G_SYSTEM_RAM:
        case S3G_SYSTEM_RAM_DYNAMIC:
            ret = s3g_map_system_ram(vma, map);
            break;

        default:
            s3g_error("%s, unknown memory map type", __func__);
            ret = -1;
            break;
    }
#endif
    
    return ret;
}

static unsigned int s3g_poll(struct file *filp, struct poll_table_struct *wait)
{
    s3g_file_t *priv = filp->private_data;

    unsigned int mask = 0; 

    poll_wait(filp, &priv->wait_queue, wait);

    if(priv->wait_events & priv->curr_events)
    {
        mask = POLLIN;
    }

    return mask;
} 

static struct file_operations s3g_fops = 
{
    .owner   = THIS_MODULE,
    .open    = s3g_open,
    .release = s3g_release,
    .poll    = s3g_poll,
    .unlocked_ioctl   = s3g_unlocked_ioctl,
#if defined(__x86_64__) && defined(HAVE_COMPAT_IOCTL)
    .compat_ioctl = s3g_unlocked_ioctl,
#endif
    .mmap = s3g_mmap
};

#ifdef __S3G_MUTEX_DEBUG__
kernel_show_stack_t kernel_show_stack;
#endif

static int __init s3g_init(void)
{
    int ret = -ENOMEM;

    ret = register_chrdev(S3G_MAJOR, "s3g", &s3g_fops);
    if(ret)
    {
        s3g_error("register_chrdev failed.\n");

        return ret;
    }

    s3g_info("S3 Graphics Co., Ltd.\n");
    s3g_info("Version: %0x.%02x.%02x%s. Build on: %s\n", DRIVER_MAJOR, DRIVER_MINOR, DRIVER_PATCHLEVEL, DRIVER_CLASS, DRIVER_DATE);

#if S3G_MALLOC_TRACK | S3G_ALLOC_PAGE_TRACK | S3G_MAP_PAGES_TRACK | S3G_MAP_IO_TRACK

    s3g_mem_track_init();

#endif

#ifdef __S3G_MUTEX_DEBUG__
    s3g_info("NOTE: It is non-offical bootleg driver for debug only, not public release driver package!");
    kernel_show_stack = (kernel_show_stack_t)kallsyms_lookup_name("show_stack");
#endif

    ret = s3g_init_driver();
    if(ret)
    {
        s3g_error("register_driver() failed in s3g_init. ret:%x.\n", ret);
    }

    return 0;
}

static void __exit s3g_exit(void)
{
    s3g_deinit_driver();
    unregister_chrdev(S3G_MAJOR, "s3g");

#if S3G_MALLOC_TRACK | S3G_ALLOC_PAGE_TRACK | S3G_MAP_PAGES_TRACK | S3G_MAP_IO_TRACK
    s3g_mem_track_list_result();
#endif

    s3g_info("exit driver.\n");
}

module_init(s3g_init);
module_exit(s3g_exit);
#ifdef __S3G_MUTEX_DEBUG__
MODULE_LICENSE("GPL");
#else
MODULE_LICENSE("S3G");
#endif
MODULE_VERSION(DRIVER_VERSION_CHAR);
#ifndef KERNEL_2_4
module_param(s3g_fb, int, 0);
module_param(s3g_swfb, int, 0);
module_param(s3g_fb_mode, charp, 0);
module_param(s3g_flip, int, 0);
#else
MODULE_PARM(s3g_fb_mode, "s");
#endif

MODULE_PARM_DESC(s3g_fb, "enable s3g fb driver");
MODULE_PARM_DESC(s3g_swfb, "Enable software fb, for embeded system fullscreen application");
MODULE_PARM_DESC(s3g_fb_mode, "Initial video mode:<xres>x<yres>-<depth>@<refresh>");
MODULE_PARM_DESC(s3g_flip, "enable frame buffer flip support");

