/*
* Copyright 2001-2015 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2012 S3 Graphics, Inc. All Rights Reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a
* copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sub license,
* and/or sell copies of the Software, and to permit persons to whom the
* Software is furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice (including the
* next paragraph) shall be included in all copies or substantial portions
* of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
* THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
* OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
* ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
* DEALINGS IN THE SOFTWARE.
*/

#include "os_interface.h"
#include "s3g_driver.h"

extern s3g_card_t s3g_cards[S3G_MAX_CARDS];
extern void  krnl_dump_resource(void *data);

int s3g_add_wait(s3g_card_t *s3g, s3g_file_t *priv, unsigned int wait_events)
{
    unsigned long flags;

    /* do not add empty wait, and one fd only can have one waiter */
    if((wait_events == 0) || (priv->wait_events != 0))
    {  
        return -1;
    }

    INIT_LIST_HEAD(&priv->wait_list);

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    priv->wait_events = wait_events;

    list_add(&priv->wait_list, &s3g->wait_list);

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);

    return 0;
}

int s3g_remove_wait(s3g_card_t *s3g, s3g_file_t *priv)
{
    unsigned long flags;

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    list_del(&priv->wait_list);

    priv->wait_events = 0;

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);

    return 0;
}

unsigned int s3g_get_event(s3g_card_t *s3g, s3g_file_t *priv)
{
    unsigned long flags;
    unsigned int  events;

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    events = priv->curr_events;
    priv->curr_events = 0;

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);

    return events;
}

int s3g_post_event(s3g_card_t *s3g, unsigned int events)
{
    s3g_file_t    *priv = NULL;
    unsigned long flags; 

    spin_lock_irqsave(&s3g->wait_list_lock, flags);

    list_for_each_entry(priv, &s3g->wait_list, wait_list)
    {
        /* only wakeup when wait event happen */
        if(events & priv->wait_events)
        {
            priv->curr_events |= events;
            wake_up_interruptible(&priv->wait_queue);
        }  
    }

    spin_unlock_irqrestore(&s3g->wait_list_lock, flags);

#ifdef NEW_S3GFB
    if(s3g->info != NULL)
    {
        if(!s3g->enter_vt)
        {
            s3gfb_handle_event(s3g, events);
        }
        else
        {
            s3g->s3gfb_pending_events |= events;
        }
    }
#endif  

    return 0;
}

pgprot_t os_get_pgprot_val(unsigned int *cache_type, pgprot_t old_prot, int io_map)
{
    pgprot_t prot = old_prot;

    if(*cache_type == S3G_MEM_UNCACHED)
    {   
        prot = pgprot_noncached(old_prot);
    }
    else if(*cache_type == S3G_MEM_WRITE_COMBINED)
    {
#ifdef CONFIG_X86
#ifdef CONFIG_X86_PAT
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,19,0)
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | cachemode2protval(_PAGE_CACHE_MODE_WC));
#else
        prot = __pgprot((pgprot_val(old_prot) & ~ _PAGE_CACHE_MASK) | _PAGE_CACHE_WC);
#endif
#else
        prot = pgprot_noncached(old_prot);

        *cache_type = S3G_MEM_UNCACHED;	
#endif
#else
        prot = pgprot_writecombine(old_prot);
#endif
        
    }
#ifdef V6_BRINGUP
    else if(*cache_type == S3G_MEM_WRITE_BACK)
    {   
        // for snoop path, current linux kennel defalut is not WA+WB
        prot = __pgprot((pgprot_val(old_prot)&(~(L_PTE_MT_MASK))) | (L_PTE_MT_WRITEBACK | L_PTE_MT_WRITEALLOC));
    }
#endif

    return prot;
}

#ifndef __frv__
int s3g_map_system_ram(struct vm_area_struct* vma, s3g_map_argu_t *map)
{
    unsigned long start = vma->vm_start;

    unsigned long i = 0, pfn;

    unsigned int  cache_type;

    struct os_pages_memory *memory = map->memory;
    
#if LINUX_VERSION_CODE <= 0x02040e
    vma->vm_flags |= VM_LOCKED;
#elif LINUX_VERSION_CODE < 0x030700
    vma->vm_flags |= VM_RESERVED;
#else
    vma->vm_flags |= VM_DONTEXPAND | VM_DONTDUMP;
#endif

    cache_type = map->flags.cache_type;

#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 0);

    for(i = 0; i < memory->page_count; i++)
    {
        pfn = page_to_pfn(memory->pages[i]);

        if(remap_pfn_range(vma, start, pfn, PAGE_SIZE, vma->vm_page_prot))
        {
            return -EAGAIN;
        }

        start += PAGE_SIZE;
    }

    return 0;
}

int s3g_map_system_io(struct vm_area_struct* vma, s3g_map_argu_t *map)
{
    unsigned int  cache_type = map->flags.cache_type;
#if LINUX_VERSION_CODE >= 0x020612
    vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
#endif
    vma->vm_page_prot = os_get_pgprot_val(&cache_type, vma->vm_page_prot, 1);

    map->flags.cache_type = cache_type;

    if (remap_pfn_range(vma, vma->vm_start,
                vma->vm_pgoff,
                vma->vm_end - vma->vm_start,
                vma->vm_page_prot))
    {
        return -EAGAIN;
    }

    return 0;    
}
#endif

static irqreturn_t s3g_isr(int irq, void *data)
{
    s3g_card_t *s3g = data;

    if((s3g->isr_func)(s3g->adapter))
    {
        tasklet_schedule(&s3g->tasklet);
    }
    
    return IRQ_HANDLED;
}

void s3g_interrupt_init(s3g_card_t *s3g)
{
    /* request isr */
    krnl_get_isr_func(s3g->adapter, (void**)&s3g->isr_func, (void**)&s3g->isr_bf_func);

    krnl_set_callback_func(s3g->adapter, OS_CALLBACK_POST_EVENT, s3g_post_event, s3g);

    tasklet_init(&s3g->tasklet, (void*)s3g->isr_bf_func, (unsigned long)s3g->adapter);
    tasklet_disable(&s3g->tasklet);
    tasklet_enable(&s3g->tasklet); 

    if(s3g_register_interrupt(s3g, s3g_isr))
    {
        s3g_error("s3g_regster_interrupt failed.\n");
    }  
}

void s3g_interrupt_reinit(s3g_card_t *s3g)
{
    /* request isr */
    krnl_get_isr_func(s3g->adapter, (void**)&s3g->isr_func, (void**)&s3g->isr_bf_func);

    krnl_set_callback_func(s3g->adapter, OS_CALLBACK_POST_EVENT, s3g_post_event, s3g);

    tasklet_enable(&s3g->tasklet); 

    if(s3g_register_interrupt(s3g, s3g_isr))
    {
        s3g_error("s3g_regster_interrupt failed.\n");
    }  
}

void s3g_interrupt_deinit(s3g_card_t *s3g)
{
    krnl_disable_interrupt(s3g->adapter);

    s3g_unregister_interrupt(s3g);

    tasklet_disable(&s3g->tasklet);
}

/***/

#include <linux/proc_fs.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)

#include <linux/seq_file.h>

static int s3g_new_read_proc(struct seq_file *m, void  *p)
{
    s3g_card_t *s3g = &s3g_cards[0];

    s3g_info("s3g_new_read_proc , s3g 0x%x\n", (unsigned int*)s3g);

    krnl_dump_resource(s3g->adapter);
    return 0;
}

static int s3g_debug_info_proc_open(struct inode *inode, struct file *file)
{
    return single_open(file ,  s3g_new_read_proc, NULL);
}

static const struct file_operations  proc_s3g_info_fops  = {
    .owner = THIS_MODULE,
    .open = s3g_debug_info_proc_open,
    .read = seq_read,
    .llseek =seq_lseek,
    .release = single_release,
};

#else
static int s3g_read_proc(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{

    s3g_card_t    *s3g = data;

    krnl_dump_resource(s3g->adapter);

    return 0;
}
#endif



int s3g_card_init(s3g_card_t *s3g, int card_num, void *pdev)
{
    int ret;
    int primary;
    unsigned short command = 0;
    unsigned int config_value = 0;

    s3g->pdev  = pdev;
    s3g->index = card_num;

    spin_lock_init(&s3g->wait_list_lock);
    INIT_LIST_HEAD(&s3g->wait_list);

    ret = s3g_init_bus_id(s3g, card_num, pdev);
    if(ret != 0)
    {
        return ret;
    }

    s3g_get_command_status16(pdev, &command);
    
    primary = (command & 0x01) ? 1 : 0;

    /* we can enable s3gfb primay and no fb driver enable by default */
    if(primary && !(s3g->vesafb_enable || s3g->vga16fb_enable))
    {
        s3g->s3gfb_enable = s3g->s3gfb_enable ? TRUE : FALSE;
    }
    else
    {
        s3g->s3gfb_enable = FALSE;
    }

    /* if fb not enable, we need reserve vmem for VGA, hardcode 256K for easy */   
    if(primary && !(s3g->s3gfb_enable || s3g->vesafb_enable || s3g->vga16fb_enable))
    {
        s3g->reserved_vmem = 256 * 1024;
        s3g_info("reserve 256K vmem for VGA.\n");
    }

    if(s3g_get_krnl_config_file_info("/etc/X11/via_kernel.conf", "freezable_patch", &config_value))
    {
        s3g->freezable_patch = config_value;
    }
    s3g_info("s3g->freezable_patch : 0x%x.\n", s3g->freezable_patch);

    s3g->adapter = krnl_init_adapter(pdev, s3g->reserved_vmem, s3g->index);

    if(s3g->adapter == NULL)
    {
        s3g_error("init adapter failed\n");
    }

    if(s3g->vesafb_enable)
    {
        krnl_notify_vesa_mode(s3g->adapter, &s3g->vesafb_mode);
    }
#ifdef S3G_PCIE_BUS
    s3g->msi_enable = 1;
#else
    s3g->msi_enable = 0;
#endif

#ifndef S3G_HW_NULL
    s3g_interrupt_init(s3g);
#endif

    s3g->lock = s3g_create_mutex();


#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
        proc_create("driver/s3g", 0644,  NULL,  &proc_s3g_info_fops);
#else
        create_proc_read_entry("driver/s3g", 0, NULL, s3g_read_proc, s3g);
#endif



    return ret;

}

int s3g_card_deinit(s3g_card_t *s3g)
{
    krnl_wait_chip_idle(s3g->adapter);

#ifndef S3G_HW_NULL
    s3g_interrupt_deinit(s3g);
#endif
    krnl_deinit_adapter(s3g->adapter);

    if(s3g->busId)
    {
        s3g_free(s3g->busId);
        s3g->busId = NULL;
        s3g->len = 0;
    }

    if(s3g->lock)
    {
        s3g_destroy_mutex(s3g->lock);
        s3g->lock = NULL;
    }

    return 0;
}

void* s3g_get_global_adapter(void)
{
    s3g_card_t *s3g = &s3g_cards[0];
    if (s3g && s3g->adapter)
    {
        return s3g->adapter;
    }
    else
    {
        s3g_error("s3g_get_adapter: s3g_card or adapter is null\n");
        return 0;
    }
}

unsigned int s3g_create_allocation_export(s3g_2d_create_allocation_t *create_data)
{
    unsigned int ret = 0;
    void *adapter = s3g_get_global_adapter();

    if (adapter)
    {
        ret = krnl_create_allocation_export(adapter, create_data);
    }

    return ret;
}

void s3g_destroy_allocation_export(unsigned int handle)
{
    void *adapter = s3g_get_global_adapter();

    if (adapter)
    {
        krnl_destroy_allocation_export(adapter, handle);
    }
}

void *s3g_lock_allocation_export(unsigned int handle, unsigned int flags)
{
    void *ret = 0;
    void *adapter = s3g_get_global_adapter();
    
    if (adapter)
    {
        ret = krnl_lock_allocation_export(adapter, handle, flags);
    }

    return ret;
}

void s3g_unlock_allocation_export(unsigned int handle)
{
    void *adapter = s3g_get_global_adapter();
    
    if (adapter)
    {
        krnl_unlock_allocation_export(adapter, handle);
    }
}

void  s3g_kickoff_dma_export(s3g_2d_draw_t *s3g_2d_draw)
{

    void *adapter = s3g_get_global_adapter();

    if (adapter)
    {
        krnl_kickoff_2d_draw(adapter, s3g_2d_draw);
    }


}

EXPORT_SYMBOL(s3g_lock_allocation_export);
EXPORT_SYMBOL(s3g_unlock_allocation_export);
EXPORT_SYMBOL(s3g_create_allocation_export);
EXPORT_SYMBOL(s3g_destroy_allocation_export);
EXPORT_SYMBOL(s3g_kickoff_dma_export);

