/*
* Copyright 2001-2015 VIA Technologies, Inc. All Rights Reserved.
* Copyright 2001-2012 S3 Graphics, Inc. All Rights Reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a
* copy of this software and associated documentation files (the "Software"),
* to deal in the Software without restriction, including without limitation
* the rights to use, copy, modify, merge, publish, distribute, sub license,
* and/or sell copies of the Software, and to permit persons to whom the
* Software is furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice (including the
* next paragraph) shall be included in all copies or substantial portions
* of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
* THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
* OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
* ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
* DEALINGS IN THE SOFTWARE.
*/

#include "s3.h"
#include "s3g_driver.h"
#include "os_interface.h"
#include "s3g_fb.h"
#include "s3g_ioctl.h"

static struct fb_info *info = NULL;
static s3g_fb_screen_t screen;
static int yoffset = -1;

/* open/release and usage marking */
static int s3g_fb_ops_open(struct fb_info* info, int user)
{
    int ret = 0;

    return ret;
}

static int s3g_fb_ops_release(struct fb_info* info, int user)
{
    int ret = 0;

    return ret;
}

/* checks var and eventually tweaks it to something supported,
* DO NOT MODIFY PAR */
static int s3g_fb_ops_check_var(struct fb_var_screeninfo* var, struct fb_info* info)
{
    s3g_mode_info_t  mode = {0};
    s3g_fb_par_t* s3g_par = info->par;
    s3g_fb_device_t* device = &s3g_par->device;
    s3g_fb_output_t* output = NULL;
    int i, mode_valid = 0;

    if (var->vmode & FB_VMODE_INTERLACED || var->vmode & FB_VMODE_DOUBLE)
    {
        return -EINVAL;
    }

    if (var->bits_per_pixel == 24)
    {
        var->bits_per_pixel = 32;
    }

    if (var->bits_per_pixel != 16 && var->bits_per_pixel != 32)
    {
        return -EINVAL;
    }

    if (var->xres > var->xres_virtual)
    {
        var->xres = var->xres_virtual;
    }

    if (var->yres > var->yres_virtual)
    {
        var->yres = var->yres_virtual;
    }

    mode.width  = var->xres;
    mode.height = var->yres;

    for(i = 0; i < device->output_count; i++)
    {
        output = &device->output[i];

        if(!output->connected)
        {
            continue;
        }

        mode_valid = s3g_fbi_mode_valid(output, &mode);

        if(mode_valid)
        {
            break;
        }
    }

    return mode_valid ? 0 : -EINVAL;
}

/* set the video mode according to info->var */
static int s3g_fb_ops_set_par(struct fb_info* info)
{
    s3g_fb_par_t* s3g_par = info->par;
    s3g_fb_device_t* device = &s3g_par->device;
    struct fb_var_screeninfo* var = &info->var;
    s3g_dpms_set_t        dpms     = {0};
    s3g_fill_allocation_t s3g_fill_data = {0};
    int ret = 0;
    int i = 0;

    /* set mode */
    for(i = 0; i < device->output_count; i++)
    {
        s3g_fb_output_t *output = &device->output[i];
        s3g_fb_crtc_t   *crtc   = output->crtc;

        if(crtc == NULL)
        {
            continue;
        }

        crtc->mode_info.crtc         = crtc->id;
        crtc->mode_info.output       = output->type;
        crtc->mode_info.bpp          = var->bits_per_pixel;
        crtc->mode_info.scale_width  =
        crtc->mode_info.width        = 
        crtc->mode_info.dst_width    = var->xres;
        crtc->mode_info.scale_height =
        crtc->mode_info.height       =
        crtc->mode_info.dst_height   = var->yres; 

        if(s3g_fbi_mode_valid(output, &crtc->mode_info))
        {
            krnl_set_mode(s3g_par->adapter, &crtc->mode_info);

            s3g_fill_data.hDevice      = s3g_par->hDevice;
            s3g_fill_data.hAllocation  = screen.hAllocation;
            s3g_fill_data.fill_pattern = 0;
            krnl_fill_allocation(s3g_par->adapter, &s3g_fill_data);

            crtc->ps_info.crtc    = crtc->id;
            crtc->ps_info.surface = screen.hAllocation;
            krnl_set_ps(s3g_par->adapter, &crtc->ps_info);		

            // power on too early , display has garbage
            // enable fb set dpms to solve switch console black screen problem
            dpms.output = output->type;
            dpms.mode   = 0;  //on
            krnl_set_dpms(s3g_par->adapter, &dpms);
        }
        else
        {
            s3g_info("fb: device: 0x%d, mode is not supported w=%d, h=%d, bpp=%d, refresh=%d!\n", 
                output->type, crtc->mode_info.width, crtc->mode_info.height, crtc->mode_info.bpp, crtc->mode_info.refresh);
        }
    }

    return ret;
}

/* set color register */
static int s3g_fb_ops_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned  blue, unsigned transp, struct fb_info* info)
{
    int ret = 0;

    if (regno >= 16)
    {
        return 1;
    }

    switch (info->var.bits_per_pixel)
    {
    case 16:
        ((unsigned int *) info->pseudo_palette)[regno] = (red & 0xf800) | ((green & 0xfc00) >> 5) |    ((blue & 0xf800) >> 11);
        break;
    case 32:
        transp >>= 8;
        red >>= 8;
        green >>= 8;
        blue >>= 8;
        ((unsigned int *) info->pseudo_palette)[regno] = (transp << 24) | (red << 16 | (green << 8) | blue);
        break;
    }

    return ret;
}

/* blank display */
static int s3g_fb_ops_blank(int blank, struct fb_info* info)
{
    int ret = 0;

    return ret;
}

/* pan display */
static int s3g_fb_ops_pan_display(struct fb_var_screeninfo* var, struct fb_info* info)
{
    int ret = 0;
    int i;
    s3g_fb_par_t* s3g_par = info->par;
    s3g_fb_device_t* device = &s3g_par->device;

    if(var->xoffset + var->xres > var->xres_virtual)
    {
        return -EINVAL;
    }

    if(var->yoffset + var->yres > var->yres_virtual)
    {
        return -EINVAL;
    }

    for (i = 0; i < device->crtc_count; i++)
    {
        if (device->dev_preferred[i]==0)
        {
            continue;
        }
#ifdef ANDROID
    {
        s3g_fb_set_ps_t set_ps = {0};
        set_ps.hAllocation = screen.hAllocation;
        set_ps.xoffset = var->xoffset;
        set_ps.yoffset = var->yoffset;
        set_ps.pipe = i;
        set_ps.pitch = screen.pitch;
        set_ps.offset = screen.fbOffset;
        set_ps.bpp = screen.bpp;
        set_ps.gbd = screen.gbd;

        s3g_par->chip_func->set_ps(s3g_par->adapter, &set_ps);
    }
#else
    {
        s3g_fb_crtc_t* crtc = device->crtc + i;
        memset(&crtc->ps_info, 0, sizeof(crtc->ps_info));
        crtc->ps_info.x = var->xoffset;
        crtc->ps_info.y = var->yoffset;
        crtc->ps_info.crtc    = i;
        crtc->ps_info.surface = screen.hAllocation;

        krnl_set_ps(s3g_par->adapter, &crtc->ps_info);
    }
#endif

    }

    return ret;
}

/* core render func may sleep, so if fb render called from a none-sleep context, this may lead error
 * so in this condition, do not use HW render, but sw instead 
 */
static int s3g_fb_can_accel(void* adapter, unsigned int ops)
{
    int can = FALSE;

    if((!in_atomic() && !irqs_disabled()) || 
        system_state != SYSTEM_RUNNING || oops_in_progress)
    {
        can = TRUE;
    }

    return can;
}

/* Draws a rectangle */
static void s3g_fb_ops_fillrect(struct fb_info* info, const struct fb_fillrect* rect)
{
    s3g_fb_par_t* s3g_par = info->par;
    void*         adapter = s3g_par->adapter;
    s3g_fillrect_t s3g_rect = {0};

    //s3g_info("fb: fill_rect(%d, %d, %d, %d)\n", rect->dx, rect->dy, rect->width, rect->height);

    if (!rect->width || !rect->height)
    {
        return;
    }

    if(!s3g_fb_can_accel(adapter, FB_OPS_FILLRECT))
    {
        cfb_fillrect(info, rect);

        return;
    }

    s3g_rect.hAllocation = screen.hAllocation;
    s3g_rect.fbOffset = screen.fbOffset;
    s3g_rect.gbd = screen.gbd;
    s3g_rect.dx = rect->dx;
    s3g_rect.dy = rect->dy;
    s3g_rect.width = rect->width;
    s3g_rect.height = rect->height;
    s3g_rect.rop = (rect->rop==ROP_XOR)?0x5a:0xcc;
    s3g_rect.color = ((unsigned int*)screen.pseudo_palette)[rect->color];

    s3g_par->chip_func->fillrect(s3g_par->adapter, &s3g_rect);
}

/* Copy data from area to another */
static void s3g_fb_ops_copyarea(struct fb_info* info, const struct fb_copyarea* region)
{
    s3g_fb_par_t* s3g_par = info->par;
    void*        adapter = s3g_par->adapter;
    s3g_copyarea_t s3g_region = {0};

    //s3g_info("fb: copy_area(%d, %d, %d, %d,   %d, %d)\n", region->dx, region->dy, region->width, region->height, region->sx, region->sy);

    if (!region->width || !region->height)
    {
        return;
    }

    if(!s3g_fb_can_accel(adapter, FB_OPS_COPYAREA))
    {
        cfb_copyarea(info, region);
        return;
    }
    s3g_region.hAllocation = screen.hAllocation;
    s3g_region.fbOffset = screen.fbOffset;
    s3g_region.pitch = screen.pitch;
    s3g_region.gbd = screen.gbd;
    s3g_region.dx = region->dx;
    s3g_region.dy = region->dy;
    s3g_region.width = region->width;
    s3g_region.height = region->height;
    s3g_region.sx = region->sx;
    s3g_region.sy = region->sy;

    s3g_par->chip_func->copyarea(s3g_par->adapter, &s3g_region);
}

/* Draws a image to the display */
static void s3g_fb_ops_imageblit(struct fb_info* info, const struct fb_image* image)
{
    s3g_fb_par_t* s3g_par;
    void*         adapter;
    s3g_image_t s3g_image = {0};

    if(!info)
    {
        return;
    }

    if(!image)
    {
        return;
    }

    s3g_par = info->par;
    if(!s3g_par)
    {
        return;
    }
    adapter = s3g_par->adapter;

    //s3g_info("fb: image_blit(%d, %d, %d, %d)\n", image->dx, image->dy, image->width, image->height);

    if (s3g_par->disabled || !image->width || !image->height)
    {
        return;
    }

    if(!s3g_fb_can_accel(adapter, FB_OPS_IMAGEBLIT))
    {
        cfb_imageblit(info, image);
        return;
    }

    s3g_image.hAllocation = screen.hAllocation;
    s3g_image.fbOffset = screen.fbOffset;
    s3g_image.gbd = screen.gbd;
    s3g_image.dx = image->dx;
    s3g_image.dy = image->dy;
    s3g_image.width = image->width;
    s3g_image.height = image->height;
    s3g_image.fg_color = ((unsigned int*)screen.pseudo_palette)[image->fg_color];
    s3g_image.bg_color = ((unsigned int*)screen.pseudo_palette)[image->bg_color];;
    s3g_image.depth = image->depth;
    s3g_image.data = image->data;

    s3g_par->chip_func->imageblit(s3g_par->adapter, &s3g_image);
}

/* wait for blit idle, optional */
static int s3g_fb_ops_sync(struct fb_info* info)
{
    s3g_fb_par_t* s3g_par = info->par;
    void*         adapter = s3g_par->adapter;

    if(s3g_fb_can_accel(adapter, FB_OPS_SYNC))
    {
        s3g_par->chip_func->sync(s3g_par->adapter);
    }

    return 0;
}

/*
 * Comment these functions for unused function warning,
 * keep the code in case it may be used in future.
 */
#if 0
static void s3g_fbi_load_cursor(struct fb_info* info, struct fb_cursor* cursor)
{
    s3g_fb_par_t* s3g_par = info->par;
    unsigned int width = (cursor->image.width + 7) & (~7);
    const char* src = cursor->image.data;
    const char* msk = cursor->mask;
    unsigned int * data;

    int i, j, k;
    unsigned char bdata;

    unsigned int fg_color = ((unsigned int *) info->pseudo_palette)[cursor->image.fg_color];
    unsigned int bg_color = ((unsigned int *) info->pseudo_palette)[cursor->image.bg_color];

    data = (unsigned int *) s3g_par->cursorVirtual;
    memset(data, 0x0, 64 * 64 * 4);
    for (j = 0; j < cursor->image.height; j++)
    {
        for (i = 0; i < width / 8; i++)
        {
            if (cursor->rop == ROP_XOR)
            {
                bdata = (*src++) ^ (*msk++);
            }
            else
            {
                bdata = (*src++) & (*msk++);
            }

            for (k = 0; k < 8; k++)
            {
                if (bdata & (1 << (8 - k - 1)))
                {
                    data[64 * j + i * 8 + k] = fg_color;
                }
                else
                {
                    data[64 * j + i * 8 + k] = bg_color;
                }
            }
        }
    }
}

/* Draws cursor */
static int s3g_fb_ops_cursor(struct fb_info* info, struct fb_cursor* cursor)
{
    s3g_fb_par_t* s3g_par = info->par;
    int crtc_count = s3g_par->device.crtc_count;
    s3g_set_cursor_t cursor_set = {0};
    int ret = 0;
    int i;

    cursor_set.hAllocation = s3g_par->hCursorAllocation;

    /* load cursor image */
    if (cursor->set & FB_CUR_SETCMAP || cursor->set & (FB_CUR_SETSHAPE | FB_CUR_SETIMAGE))
    {
        s3g_fbi_load_cursor(info, cursor);
    }

    /* hide cursor */
    for (i = 0; i < crtc_count; i++)
    {
        cursor_set.crtc   = i;
        cursor_set.opcode = S3G_CURSOR_HIDE;

        krnl_set_cursor(s3g_par->adapter, &cursor_set);
    }

    /* set cursor position */
    if (cursor->set & FB_CUR_SETPOS)
    {
        cursor_set.opcode = S3G_CURSOR_SET_POS;
        cursor_set.x = cursor->image.dx - info->var.xoffset;
        cursor_set.y = cursor->image.dy - info->var.yoffset;

        for (i = 0; i < crtc_count; i++)
        {
            cursor_set.crtc = i;
            krnl_set_cursor(s3g_par->adapter, &cursor_set);
        }
    }

    /* show cursor */
    if (cursor->enable)
    {
        cursor_set.opcode = S3G_CURSOR_SHOW;

        for (i = 0; i < crtc_count; i++)
        {
            cursor_set.crtc = i;
            krnl_set_cursor(s3g_par->adapter, &cursor_set);
        }
    }

    return ret;
}
#endif

static void s3g_fb_private_load_cursor(s3g_fb_par_t *s3g_par)
{
    int i, j;

    s3g_create_allocation_t create_data = {0};

    s3g_memset(&create_data, 0, sizeof(create_data));

    create_data.width  = 64;
    create_data.height = 64;
    create_data.tiled  = 0;
    create_data.access_hint = S3G_ACCESS_GPU_ALMOST;
    create_data.usage_mask  = S3G_USAGE_CURSOR_SURFACE;
    create_data.format      = S3G_FORMAT_B8G8R8A8_UNORM;
    create_data.unpagable   = 1;
    create_data.device      = s3g_par->hDevice;

    krnl_create_allocation(s3g_par->adapter, &create_data);

    s3g_par->hPrivCursorAllocation     = create_data.allocation;
    s3g_par->privCursorVirtual         = s3g_fbi_map(s3g_par, s3g_par->hPrivCursorAllocation);

    /* load cursor image */

    if(s3g_par->privCursorVirtual)
    {
        unsigned int *data;

        memset(s3g_par->privCursorVirtual, 0, create_data.pitch * create_data.height);

        for(j = 0; j < 16; j++)
        {
            for(i = 0; i < 16; i++)
            {         
                data = (unsigned int*)((char*)s3g_par->privCursorVirtual + create_data.pitch * j) + i;
                if((i-8)*(i-8) + (j-8)*(j-8) <= 64)
                {
                    *data = 0xffffffff;
                }
            }
        }

        s3g_fbi_unmap(s3g_par, s3g_par->hPrivCursorAllocation);
        s3g_par->privCursorVirtual = NULL;    
    }    
}

static void s3g_fb_destroy_priv_cursor(s3g_fb_par_t *s3g_par)
{

    if (s3g_par->hPrivCursorAllocation)
    {
        krnl_destroy_2d_allocation(s3g_par->adapter, s3g_par->hPrivCursorAllocation);
        s3g_par->hPrivCursorAllocation = 0;
    }


}

static void s3g_fb_private_set_cursor(s3g_fb_par_t *s3g_par, s3g_set_cursor_t *cursor_set)
{
    int i;
    int crtc_count = s3g_par->device.crtc_count;

    cursor_set->hAllocation = s3g_par->hPrivCursorAllocation;

    if(cursor_set->opcode == S3G_CURSOR_SHOW)
    {
        /* load cursor */
        s3g_fb_private_load_cursor(s3g_par);
    }

    /* show curosr */
    for (i = 0; i < crtc_count; i++)
    {
        cursor_set->crtc = i;

        krnl_set_cursor(s3g_par->adapter, cursor_set);
    }
}

static int s3g_fb_private_set_rotation(s3g_fb_par_t *s3g_par, unsigned char rotation_degree)
{
    int ret = 0;

    //to do: when spec ready

    return ret;
}

/* perform fb specific ioctl (optional) */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15) 
static int s3g_fb_ops_ioctl(struct inode *inode, struct file *file, unsigned int cmd,unsigned long arg, struct fb_info *info)
#else
static int s3g_fb_ops_ioctl(struct fb_info* info, unsigned int cmd, unsigned long arg)
#endif
{
    int ret = 0;

    switch(cmd)
    {
    case FB_S3G_IOCTL_GET_ALLOCATION:
        if(copy_to_user((void*)arg, &screen.hAllocation, sizeof(unsigned int)))
        {
            return -1;
        }
        break;

    case FB_S3G_IOCTL_SET_CURSOR:
    {
        s3g_set_cursor_t cursor;

        if(copy_from_user((void*)&cursor, (void*)arg, sizeof(cursor)))
        {
            return -1;
        }

        s3g_fb_private_set_cursor(info->par, &cursor);

        break;
    }

    /*QUEUE_POST_ALLOCATION :
     * Recode the yoffset of the allocation,which need to be show*/
    case FB_S3G_IOCTL_QUEUE_POST_ALLOCATION:
    {
        if(copy_from_user((void*)&yoffset, (void*)arg, sizeof(yoffset)))
        {
            return -1;
        }
        break;
    }

    /*DEQUEUE_POST_ALLOCATION:
     * Get the yoffset of the allocation, which need to be show.*/
    case FB_S3G_IOCTL_DEQUEUE_POST_ALLOCATION:
    {
        if(copy_to_user((void*)arg, &yoffset, sizeof(int)))
        {
            return -1;
        }
        
        yoffset = -1;

        break;
    }

    case FB_S3G_IOCTL_GET_FRAMEBUFFER_INFO:
    {
        s3g_fb_info_t fb_info;

        fb_info.width = screen.width;
        fb_info.height = screen.height;
        fb_info.pitch = screen.pitch;
        fb_info.bpp = screen.bpp;
        fb_info.gbd = screen.gbd;
        fb_info.hallocation = screen.hAllocation;
        fb_info.offset = screen.fbOffset;

        if(copy_to_user((void*)arg, &fb_info, sizeof(fb_info)))
        {
            return -1;
        }

        break;
    }

    //case FB_S3G_IOCTL_SET_ROTATION:
    case 6:
    {

        ret = s3g_fb_private_set_rotation(info->par, *(unsigned char*)arg);

        break;
    }

    default:
        ret = -1;
        break;
    }

    return ret;
}

static int s3g_fb_ops_mmap(struct fb_info *info, struct vm_area_struct *vma)
{
    unsigned long off;
    unsigned long phys_addr;
    unsigned long size;

    int ret = -EINVAL;

    int mem_type;

    s3g_fb_par_t* s3g_par = info->par;

    s3g_map_argu_t map;

    struct os_pages_memory *pages = NULL;
    alloc_pages_flags_t alloc_flags = {0};

    if(vma->vm_end - vma->vm_start == 0)
        return 0;

    size = vma->vm_end - vma->vm_start;

    if(vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
        return -EINVAL;

    off = vma->vm_pgoff << PAGE_SHIFT;

    phys_addr = off + screen.fbOffset;

    mem_type = s3g_par->chip_func->get_local_mem_type();

    if(mem_type == S3G_SYSTEM_RAM_DYNAMIC)
    {
        alloc_flags.continuous = 1;

        pages = s3g_allocate_pages_memory_dynamic(size, alloc_flags);

        krnl_get_pgtbl_entry_dynamic(s3g_par->adapter, phys_addr, size, pages);

        s3g_memset (&map, 0, sizeof (s3g_map_argu_t));
        map.flags.cache_type = S3G_MEM_WRITE_COMBINED;
        map.memory = pages;
        map.flags.mem_type = mem_type;
        map.flags.mem_space = S3G_MEM_USER;
        map.size = size;

        ret = s3g_map_system_ram(vma, &map);

        s3g_free_pages_memory_dynamic(pages);
    }
    else
    {
        ret = -EINVAL;
        s3g_info("error fb map !\n");
    }

    return ret;
}


static struct fb_ops s3g_fb_ops =
{
    .owner          = THIS_MODULE,
    .fb_open        = s3g_fb_ops_open,
    .fb_release     = s3g_fb_ops_release,
    .fb_check_var   = s3g_fb_ops_check_var,
    .fb_set_par     = s3g_fb_ops_set_par,
    .fb_setcolreg   = s3g_fb_ops_setcolreg,
    .fb_blank       = s3g_fb_ops_blank,
    .fb_pan_display = s3g_fb_ops_pan_display,
    .fb_fillrect    = s3g_fb_ops_fillrect,
    .fb_copyarea    = s3g_fb_ops_copyarea,
    .fb_imageblit   = s3g_fb_ops_imageblit,
    .fb_ioctl       = s3g_fb_ops_ioctl,
    .fb_cursor      = NULL,                  //s3g_fb_ops_cursor, just comment it
    .fb_sync        = s3g_fb_ops_sync,
};

int s3g_fb_probe(s3g_card_t *s3g, char* mode_string, int flip)
{
    s3g_fb_mode_t init_mode = {0};
    int ret = 0;
    void *adapter = s3g->adapter;
    void *pdev = s3g->pdev;
    unsigned int fbBase = 0;
    unsigned int screenSize = 0;
    unsigned int mmioBase = 0;
    unsigned int mmioSize = 0;
    int  height_alignment, alignedheight;
    int  mem_type;

    bus_config_t  bus_config = {0};
    s3g_query_info_t query = {0};
    s3g_2d_screen_t create_screen = {0};
    s3g_create_allocation_t create_data = {0};
    s3g_fb_par_t* s3g_par = NULL;

    s3g_get_bus_config(s3g->pdev, &bus_config);

    mmioBase = bus_config.start_addr[0];
    mmioSize = bus_config.end_addr[0] -bus_config.start_addr[0] + 1;
    fbBase = bus_config.start_addr[1];

    /* create fb_info */
#ifdef S3G_PCIE_BUS
    info = framebuffer_alloc(sizeof(s3g_fb_par_t), &((struct pci_dev*)pdev)->dev);
#else
    info = framebuffer_alloc(sizeof(s3g_fb_par_t), &((struct platform_device*)pdev)->dev);
#endif

    s3g_par = info->par;
    s3g_par->adapter = adapter;

    query.type = S3G_QUERY_CHIP_ID;

    krnl_query_info(adapter, &query);

    /* init chip function pointer */
    switch ( query.value )
    {
    case CHIP_DST2:
    case CHIP_DST3:
    case CHIP_DST4:
    case CHIP_DUMA:
        s3g_par->chip_func = &s3g_fb_function_dst;
        break;
    case CHIP_ELT:
    case CHIP_ELT1K:
        s3g_par->chip_func = &s3g_fb_function_elt;
        break;
    case CHIP_E2UMA:
        s3g_par->chip_func = &s3g_fb_function_exc;
        break;
    default:
        s3g_par->chip_func = NULL;
        break;
    }

#ifndef S3G_HW_NULL
    /* init device info */
    if(s3g_fbi_init_device(adapter, &s3g_par->device))
    {
        framebuffer_release(info);
        return -1;
    }
#endif

    /* parse mode info */
    s3g_fbi_parse_screen_mode(adapter, mode_string, &init_mode);

#ifndef S3G_HW_NULL
    if(s3g_fbi_fix_screen_mode(adapter, &s3g_par->device, &init_mode))
    {
        s3g_fbi_deinit_device(adapter, &s3g_par->device);
        framebuffer_release(info);
        return -1;
    }

    s3g_fbi_init_crtc_mode_set(adapter, &s3g_par->device, &init_mode);
#endif

    /* Create 2d device */
    s3g_par->hDevice = krnl_create_2ddevice(adapter, NULL); 

    height_alignment = s3g_par->chip_func->getheightalign(init_mode.bpp);
    alignedheight = (init_mode.height + height_alignment - 1) & (~(height_alignment-1));
    create_screen.width  = init_mode.width;
    create_screen.height = flip ? alignedheight*2 : alignedheight;
    create_screen.bpp    = init_mode.bpp;
    create_screen.device = s3g_par->hDevice;

    s3g_init_2d_screen(s3g, &create_screen);

    screenSize = create_screen.pitch * create_screen.height;

    screen.width  = create_screen.width;
    screen.height = create_screen.height;
    screen.bpp    = create_screen.bpp;
    screen.pitch  = create_screen.pitch;
    screen.gbd    = create_screen.gbd;
    screen.hAllocation = create_screen.hAllocation;
    screen.fbOffset    = create_screen.fbOffset;
    screen.fbVirtual   = s3g_fbi_map(s3g_par, screen.hAllocation);

    /* create cursor */
    s3g_memset(&create_data, 0, sizeof(create_data));

    create_data.width  = 64;
    create_data.height = 64;
    create_data.tiled  = 0;
    create_data.access_hint = S3G_ACCESS_GPU_ALMOST;
    create_data.usage_mask  = S3G_USAGE_CURSOR_SURFACE;
    create_data.format      = S3G_FORMAT_B8G8R8A8_UNORM;
    create_data.unpagable   = 1;
    create_data.device      = s3g_par->hDevice;

    krnl_create_allocation(adapter, &create_data);

    s3g_par->hCursorAllocation = create_data.allocation;
    s3g_par->cursorVirtual     = s3g_fbi_map(s3g_par, s3g_par->hCursorAllocation);

    /* fill fb info struct */
    mem_type = s3g_par->chip_func->get_local_mem_type();

    if(mem_type == S3G_SYSTEM_RAM_DYNAMIC)
    {
        s3g_fb_ops.fb_mmap = s3g_fb_ops_mmap;
    }

    if(0 != s3g_fbi_get_moniter_physical_size(adapter, &s3g_par->device, &init_mode, &query))
    {
        query.MonitorHorSize = init_mode.dstWidth;
        query.MonitorVerSize = init_mode.dstHeight;
        
        s3g_info("get device physical size fail, use default setting.\n ");
    }

    s3g_info("finally, fixed monitor size is [%d, %d](mm)", query.MonitorHorSize, query.MonitorVerSize);
    
    info->screen_base = screen.fbVirtual;
    info->fbops = &s3g_fb_ops;
    info->screen_size = screenSize;
    info->pseudo_palette = screen.pseudo_palette;

    info->flags = FBINFO_DEFAULT | (FBINFO_HWACCEL_COPYAREA | FBINFO_HWACCEL_FILLRECT | FBINFO_HWACCEL_IMAGEBLIT);

    info->var.xres_virtual = screen.width;
    info->var.yres_virtual = screen.height;
    info->var.bits_per_pixel = screen.bpp;
    info->var.xoffset = 0;
    info->var.yoffset = 0;
    info->var.activate = FB_ACTIVATE_NOW;
    info->var.height   = query.MonitorVerSize;
    info->var.width    = query.MonitorHorSize;
    info->var.xres     = init_mode.width;
    info->var.yres     = init_mode.height;

    info->var.pixclock = -1;

    switch (init_mode.bpp)
    {
    case 16:
        info->var.red.offset = 11;
        info->var.green.offset = 5;
        info->var.blue.offset = 0;
        info->var.red.length = 5;
        info->var.green.length = 6;
        info->var.blue.length = 5;
        break;
    case 32:
        info->var.red.offset = 16;
        info->var.green.offset = 8;
        info->var.blue.offset = 0;
        info->var.red.length = 8;
        info->var.green.length = 8;
        info->var.blue.length = 8;
        break;
    }

    s3g_vsprintf(info->fix.id, "%s", "s3gfb");
    info->fix.type = FB_TYPE_PACKED_PIXELS;
    info->fix.visual = FB_VISUAL_TRUECOLOR;
    info->fix.type_aux = 0;
    info->fix.xpanstep = 1;
    info->fix.ypanstep = 1;
    info->fix.ywrapstep = 0;
    info->fix.accel = 77;
    info->fix.line_length = screen.pitch;
    
    info->fix.smem_start = fbBase + screen.fbOffset;
    info->fix.smem_len = screenSize;
    info->fix.mmio_start = mmioBase;
    info->fix.mmio_len = mmioSize;

    s3g_par->fbBase = fbBase;
    s3g->info = info;

#ifndef S3G_HW_NULL
    ret = register_framebuffer(info);
#endif

    return ret;
}

int s3g_fb_shutdown(s3g_card_t *s3g)
{
    s3g_fb_par_t* s3g_par = info->par;

    s3g_par->disabled = 1;

    return 0;
}

int s3g_fb_remove(s3g_card_t *s3g)
{
    s3g_fb_par_t* s3g_par = info->par;

    /* need power off crtc first, then do unregister fb, it can fix zhong-ke-fang-de v1.5
        CRT display will appear noise before system restart bug */
    krnl_power_off_crtc(s3g_par->adapter);

    if (info)
    {
        info->par = NULL;
#ifndef S3G_HW_NULL
        unregister_framebuffer(info);
        framebuffer_release(info);
#endif
        info = NULL;
    }

#ifndef S3G_HW_NULL
    s3g_fbi_deinit_device(s3g_par->adapter, &s3g_par->device);

    s3g_fb_destroy_priv_cursor(s3g_par);
#endif

    if (s3g_par->cursorVirtual)
    {
        s3g_fbi_unmap(s3g_par, s3g_par->hCursorAllocation);

        s3g_par->cursorVirtual = 0;
    }

    if (s3g_par->hCursorAllocation)
    {        
        krnl_destroy_2d_allocation(s3g_par->adapter, s3g_par->hCursorAllocation);

        s3g_par->hCursorAllocation = 0;
    }

    if(screen.fbVirtual)
    {
        s3g_fbi_unmap(s3g_par, screen.hAllocation);

        screen.fbVirtual = 0;
    }

    if (screen.hAllocation)
    {
        s3g_2d_screen_t screen_2d = {0};
        
        s3g_deinit_2d_screen(s3g, &screen_2d);

        screen.hAllocation = 0;
    }

    if (s3g_par->adapter)
    {
        krnl_destroy_2ddevice(s3g_par->adapter, NULL);
        s3g_par->adapter = NULL;
    }

    return 0;
}

void s3g_fb_set_pm(s3g_card_t *s3g, int bsuspend)
{
    struct fb_info *info = s3g->info;
    
    if(info == NULL)
    {
        return;
    }
    
    acquire_console_sem();
    if(bsuspend)
    {
        fb_set_suspend(info, 1);
    }
    else
    {
        fb_set_suspend(info, 0);
    }
    release_console_sem();
}

#define VESA_MEM_ALIGN      64*1024

void s3g_fb_detect_vesa(s3g_card_t *s3g, void *pdev)
{
    struct fb_info   *vesafb    = NULL;
    bus_config_t     bus_config = {0};

    int i, vesafb_found = 0, vga16fb_found = 0;

    s3g_get_bus_config(pdev, &bus_config);

    for(i = 0; i < num_registered_fb; i++)
    {
        vesafb = registered_fb[i];

        if((vesafb != NULL) && 
           (vesafb->fix.smem_start >= bus_config.start_addr[1]) && 
           (vesafb->fix.smem_start <  bus_config.end_addr[1]))
        {
            vesafb_found = 1;
            break;
        } 

        if((vesafb != NULL) && 
           (vesafb->fix.smem_start == 0xA0000))
        {
            vga16fb_found = 1;
            break;
        }
    }

    if(s3g->s3gfb_enable && (vga16fb_found || vesafb_found))
    {
        s3g_info("%s detected, conflicted with s3gfb, remove %s.\n", vesafb->fix.id, vesafb->fix.id);

        unregister_framebuffer(vesafb);
        release_mem_region(vesafb->fix.smem_start, screen_info.lfb_size * 65536);

        vga16fb_found = 0;
        vesafb_found  = 0;
    }

    if(vga16fb_found)
    {
         s3g->vga16fb_enable = 1;
         s3g->reserved_vmem  = 256 * 1024;

         s3g_info("vga16fb detected, reserved %dK vmem\n", s3g->reserved_vmem/1024);
    }

    if(vesafb_found)
    {
        s3g->vesafb_mode.crtc       = 0;
        s3g->vesafb_mode.output     = 0;
        s3g->vesafb_mode.bpp        = vesafb->var.bits_per_pixel;
        s3g->vesafb_mode.width      = vesafb->var.xres;
        s3g->vesafb_mode.height     = vesafb->var.yres;
        s3g->vesafb_mode.refresh    = 0;
        s3g->vesafb_mode.dst_width   = vesafb->var.xres;
        s3g->vesafb_mode.dst_height  = vesafb->var.yres;
        s3g->vesafb_mode.dst_refresh = 0;
        s3g->vesafb_mode.scaler_type = 0;
        s3g->vesafb_mode.interlace   = 0;
        s3g->vesafb_mode.scale_width  = vesafb->var.xres;
        s3g->vesafb_mode.scale_height = vesafb->var.yres;

        s3g->vesafb_enable = 1;
        s3g->reserved_vmem = (vesafb->fix.smem_len + VESA_MEM_ALIGN - 1) & (~(VESA_MEM_ALIGN-1));

        s3g_info("vesafb detected, %dx%d@%d, reserved %dK video memory for it\n", 
             vesafb->var.xres, vesafb->var.yres, vesafb->var.bits_per_pixel,s3g->reserved_vmem/1024);
    }
}
